#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Autowaf, useful waf utilities with support for recursive projects
# Copyright 2008-2011 David Robillard
#
# Licensed under the GNU GPL v2 or later, see COPYING file for details.

import glob
import os
import subprocess
import sys
import shutil

from waflib import Configure, Context, Logs, Node, Options, Task, Utils
from waflib.TaskGen import feature, before, after

global g_is_child
g_is_child = False

# Only run autowaf hooks once (even if sub projects call several times)
global g_step
g_step = 0

# Compute dependencies globally
#import preproc
#preproc.go_absolute = True

@feature('c', 'cxx')
@after('apply_incpaths')
def include_config_h(self):
    self.env.append_value('INCPATHS', self.bld.bldnode.abspath())

def set_options(opt, debug_by_default=False):
    "Add standard autowaf options if they haven't been added yet"
    global g_step
    if g_step > 0:
        return

    # Install directory options
    dirs_options = opt.add_option_group('Installation directories', '')

    # Move --prefix and --destdir to directory options group
    for k in ('--prefix', '--destdir'):
        option = opt.parser.get_option(k)
        if option:
            opt.parser.remove_option(k)
            dirs_options.add_option(option)

    # Standard directory options
    dirs_options.add_option('--bindir', type='string',
                            help="Executable programs [Default: PREFIX/bin]")
    dirs_options.add_option('--configdir', type='string',
                            help="Configuration data [Default: PREFIX/etc]")
    dirs_options.add_option('--datadir', type='string',
                            help="Shared data [Default: PREFIX/share]")
    dirs_options.add_option('--includedir', type='string',
                            help="Header files [Default: PREFIX/include]")
    dirs_options.add_option('--libdir', type='string',
                            help="Libraries [Default: PREFIX/lib]")
    dirs_options.add_option('--mandir', type='string',
                            help="Manual pages [Default: DATADIR/man]")
    dirs_options.add_option('--docdir', type='string',
                            help="HTML documentation [Default: DATADIR/doc]")

    # Build options
    if debug_by_default:
        opt.add_option('--optimize', action='store_false', default=True, dest='debug',
                       help="Build optimized binaries (not debug)")
    else:
        opt.add_option('--debug', action='store_true', default=False, dest='debug',
                       help="Build debuggable binaries (not optimized)")

    opt.add_option('--pardebug', action='store_true', default=False, dest='pardebug',
                       help="Build parallel-installable debuggable libraries with D suffix")

    opt.add_option('--grind', action='store_true', default=False, dest='grind',
                   help="Run tests in valgrind")
    opt.add_option('--strict', action='store_true', default=False, dest='strict',
                   help="Use strict compiler flags and show all warnings")
    opt.add_option('--ultra-strict', action='store_true', default=False, dest='ultra_strict',
                   help="Use even stricter compiler flags (likely to trigger many warnings in library headers)")
    opt.add_option('--docs', action='store_true', default=False, dest='docs',
                   help="Build documentation - requires doxygen")

    g_step = 1

def copyfile (task):
    # a cross-platform utility for copying files as part of tasks
    src = task.inputs[0].abspath()
    tgt = task.outputs[0].abspath()
    shutil.copy2 (src, tgt)

def check_header(conf, lang, name, define='', mandatory=True):
    "Check for a header"
    includes = '' # search default system include paths
    if sys.platform == "darwin":
        includes = '/opt/local/include'

    if lang == 'c':
        check_func = conf.check_cc
    elif lang == 'cxx':
        check_func = conf.check_cxx
    else:
        Logs.error("Unknown header language `%s'" % lang)
        return

    if define != '':
        check_func(header_name=name, includes=includes,
                   define_name=define, mandatory=mandatory)
    else:
        check_func(header_name=name, includes=includes, mandatory=mandatory)

def nameify(name):
    return name.replace('/', '_').replace('++', 'PP').replace('-', '_').replace('.', '_')

def define(conf, var_name, value):
    conf.define(var_name, value)
    conf.env[var_name] = value

def check_pkg(conf, name, **args):
    "Check for a package iff it hasn't been checked for yet"
    if args['uselib_store'].lower() in conf.env['AUTOWAF_LOCAL_LIBS']:
        return
    class CheckType:
        OPTIONAL=1
        MANDATORY=2
    var_name = 'CHECKED_' + nameify(args['uselib_store'])
    check = not var_name in conf.env
    mandatory = not 'mandatory' in args or args['mandatory']
    if not check and 'atleast_version' in args:
        # Re-check if version is newer than previous check
        checked_version = conf.env['VERSION_' + name]
        if checked_version and checked_version < args['atleast_version']:
            check = True;
    if not check and mandatory and conf.env[var_name] == CheckType.OPTIONAL:
        # Re-check if previous check was optional but this one is mandatory
        check = True;
    if check:
        found = None
        pkg_var_name = 'PKG_' + name.replace('-', '_')
        pkg_name = name
        if mandatory:
            args['mandatory'] = True  # Unsmash mandatory arg
        if 'atleast_version' in args:
            if not 'msg' in args:
                args['msg'] = 'Checking for %r >= %s' %(pkg_name, args['atleast_version'])
            found = conf.check_cfg(package=pkg_name, args=[pkg_name + " >= " + args['atleast_version'], '--cflags', '--libs'], **args)
        else:
            found = conf.check_cfg(package=pkg_name, args="--cflags --libs", **args)
        if found:
            conf.env[pkg_var_name] = pkg_name
        if 'atleast_version' in args:
            conf.env['VERSION_' + name] = args['atleast_version']
    if mandatory:
        conf.env[var_name] = CheckType.MANDATORY
    else:
        conf.env[var_name] = CheckType.OPTIONAL


def normpath(path):
    if sys.platform == 'win32':
        return os.path.normpath(path).replace('\\', '/')
    else:
        return os.path.normpath(path)

def ensure_visible_symbols(bld, visible):
    if bld.env['MSVC_COMPILER']:
        if visible:
            print ('*** WARNING: MSVC does not allow symbols to be visible/exported by default while building ' + bld.name)
        else:
            pass
    else:
        if not hasattr (bld,'cxxflags'):
            bld.cxxflags = []
        if not hasattr (bld,'cflags'):
            bld.cflags = []
        if visible:
            bld.cxxflags += [ '-fvisibility=default' ]
            bld.cflags += [ '-fvisibility=default' ]
        else:
            bld.cxxflags += [ '-fvisibility=hidden' ]
            bld.cflags += [ '-fvisibility=hidden' ]

def set_basic_compiler_flags(conf, flag_dict):
    if Options.options.debug:
        conf.env.append_value('CFLAGS', flag_dict['debuggable'])
        conf.env.append_value('CXXFLAGS', flag_dict['debuggable'])
        conf.env.append_value('LINKFLAGS', flag_dict['linker-debuggable'])
    else:
        conf.env.append_value('CFLAGS', flag_dict['nondebuggable'])
        conf.env.append_value('CXXFLAGS', flag_dict['nondebuggable'])

    if Options.options.ultra_strict:
        Options.options.strict = True
        conf.env.append_value('CFLAGS', flag_dict['ultra-strict'])

    if Options.options.strict:
        conf.env.append_value('CFLAGS', flag_dict['c-strict'])
        conf.env.append_value('CXXFLAGS', flag_dict['cxx-strict'])
        conf.env.append_value('CFLAGS', flag_dict['strict'])
        conf.env.append_value('CXXFLAGS', flag_dict['strict'])

    conf.env.append_value('CFLAGS', flag_dict['show-column'])
    conf.env.append_value('CXXFLAGS', flag_dict['show-column'])

def configure(conf):
    global g_step
    if g_step > 1:
        return
    print('')
    display_header('Global Configuration')

    if Options.options.docs:
        conf.load('doxygen')

    conf.env['DOCS'] = Options.options.docs
    conf.env['DEBUG'] = Options.options.debug or Options.options.pardebug
    conf.env['PARDEBUG'] = Options.options.pardebug
    conf.env['PREFIX'] = normpath(os.path.abspath(os.path.expanduser(conf.env['PREFIX'])))

    def config_dir(var, opt, default):
        if opt:
            conf.env[var] = normpath(opt)
        else:
            conf.env[var] = normpath(default)

    opts   = Options.options
    prefix = conf.env['PREFIX']

    config_dir('BINDIR',     opts.bindir,     os.path.join(prefix, 'bin'))
    config_dir('SYSCONFDIR', opts.configdir,  os.path.join(prefix, 'etc'))
    config_dir('DATADIR',    opts.datadir,    os.path.join(prefix, 'share'))
    config_dir('INCLUDEDIR', opts.includedir, os.path.join(prefix, 'include'))
    config_dir('LIBDIR',     opts.libdir,     os.path.join(prefix, 'lib'))
    config_dir('MANDIR',     opts.mandir,     os.path.join(conf.env['DATADIR'], 'man'))
    config_dir('DOCDIR',     opts.docdir,     os.path.join(conf.env['DATADIR'], 'doc'))

    if Options.options.docs:
        doxygen = conf.find_program('doxygen')
        if not doxygen:
            conf.fatal("Doxygen is required to build with --docs")

        dot = conf.find_program('dot')
        if not dot:
            conf.fatal("Graphviz (dot) is required to build with --docs")

    conf.env.prepend_value('CFLAGS', '-I' + os.path.abspath('.'))
    conf.env.prepend_value('CXXFLAGS', '-I' + os.path.abspath('.'))

    display_msg(conf, "Install prefix", conf.env['PREFIX'])
    display_msg(conf, "Debuggable build", str(conf.env['DEBUG']))
    display_msg(conf, "Build documentation", str(conf.env['DOCS']))
    print('')

    g_step = 2

def set_local_lib(conf, name, has_objects):
    var_name = 'HAVE_' + nameify(name.upper())
    define(conf, var_name, 1)
    if has_objects:
        if type(conf.env['AUTOWAF_LOCAL_LIBS']) != dict:
            conf.env['AUTOWAF_LOCAL_LIBS'] = {}
        conf.env['AUTOWAF_LOCAL_LIBS'][name.lower()] = True
    else:
        if type(conf.env['AUTOWAF_LOCAL_HEADERS']) != dict:
            conf.env['AUTOWAF_LOCAL_HEADERS'] = {}
        conf.env['AUTOWAF_LOCAL_HEADERS'][name.lower()] = True

def append_property(obj, key, val):
    if hasattr(obj, key):
        setattr(obj, key, getattr(obj, key) + val)
    else:
        setattr(obj, key, val)

def use_lib(bld, obj, libs):
    abssrcdir = os.path.abspath('.')
    libs_list = libs.split()
    for l in libs_list:
        in_headers = l.lower() in bld.env['AUTOWAF_LOCAL_HEADERS']
        in_libs    = l.lower() in bld.env['AUTOWAF_LOCAL_LIBS']
        if in_libs:
            append_property(obj, 'use', ' lib%s ' % l.lower())
            append_property(obj, 'framework', bld.env['FRAMEWORK_' + l])
        if in_headers or in_libs:
            inc_flag = '-iquote ' + os.path.join(abssrcdir, l.lower())
            for f in ['CFLAGS', 'CXXFLAGS']:
                if not inc_flag in bld.env[f]:
                    bld.env.prepend_value(f, inc_flag)
        else:
            append_property(obj, 'uselib', ' ' + l)

@feature('c', 'cxx')
@before('apply_link')
def version_lib(self):
    if sys.platform == 'win32':
        self.vnum = None  # Prevent waf from automatically appending -0
    if self.env['PARDEBUG']:
        applicable = ['cshlib', 'cxxshlib', 'cstlib', 'cxxstlib']
        if [x for x in applicable if x in self.features]:
            self.target = self.target + 'D'

def set_lib_env(conf, name, version):
    'Set up environment for local library as if found via pkg-config.'
    NAME         = name.upper()
    major_ver    = version.split('.')[0]
    pkg_var_name = 'PKG_' + name.replace('-', '_')
    lib_name     = '%s-%s' % (name, major_ver)
    if conf.env.PARDEBUG:
        lib_name += 'D'
    conf.env[pkg_var_name]       = lib_name
    conf.env['INCLUDES_' + NAME] = ['${INCLUDEDIR}/%s-%s' % (name, major_ver)]
    conf.env['LIBPATH_' + NAME]  = [conf.env.LIBDIR]
    conf.env['LIB_' + NAME]      = [lib_name]

def display_header(title):
    Logs.pprint('BOLD', title)

def display_msg(conf, msg, status = None, color = None):
    color = 'CYAN'
    if type(status) == bool and status or status == "True":
        color = 'GREEN'
    elif type(status) == bool and not status or status == "False":
        color = 'YELLOW'
    Logs.pprint('BOLD', " *", sep='')
    Logs.pprint('NORMAL', "%s" % msg.ljust(conf.line_just - 3), sep='')
    Logs.pprint('BOLD', ":", sep='')
    Logs.pprint(color, status)

def link_flags(env, lib):
    return ' '.join(map(lambda x: env['LIB_ST'] % x, env['LIB_' + lib]))

def compile_flags(env, lib):
    return ' '.join(map(lambda x: env['CPPPATH_ST'] % x, env['INCLUDES_' + lib]))

def set_recursive():
    global g_is_child
    g_is_child = True

def is_child():
    global g_is_child
    return g_is_child

# Pkg-config file
def build_pc(bld, name, version, version_suffix, libs, subst_dict={}):
    '''Build a pkg-config file for a library.
    name           -- uppercase variable name     (e.g. 'SOMENAME')
    version        -- version string              (e.g. '1.2.3')
    version_suffix -- name version suffix         (e.g. '2')
    libs           -- string/list of dependencies (e.g. 'LIBFOO GLIB')
    '''
    pkg_prefix       = bld.env['PREFIX']
    if pkg_prefix[-1] == '/':
        pkg_prefix = pkg_prefix[:-1]

    target = name.lower()
    if version_suffix != '':
        target += '-' + version_suffix

    if bld.env['PARDEBUG']:
        target += 'D'

    target += '.pc'

    libdir = bld.env['LIBDIR']
    if libdir.startswith(pkg_prefix):
        libdir = libdir.replace(pkg_prefix, '${exec_prefix}')

    includedir = bld.env['INCLUDEDIR']
    if includedir.startswith(pkg_prefix):
        includedir = includedir.replace(pkg_prefix, '${prefix}')

    obj = bld(features     = 'subst',
              source       = '%s.pc.in' % name.lower(),
              target       = target,
              install_path = os.path.join(bld.env['LIBDIR'], 'pkgconfig'),
              exec_prefix  = '${prefix}',
              PREFIX       = pkg_prefix,
              EXEC_PREFIX  = '${prefix}',
              LIBDIR       = libdir,
              INCLUDEDIR   = includedir)

    if type(libs) != list:
        libs = libs.split()

    subst_dict[name + '_VERSION'] = version
    subst_dict[name + '_MAJOR_VERSION'] = version[0:version.find('.')]
    for i in libs:
        subst_dict[i + '_LIBS']   = link_flags(bld.env, i)
        lib_cflags = compile_flags(bld.env, i)
        if lib_cflags == '':
            lib_cflags = ' '
        subst_dict[i + '_CFLAGS'] = lib_cflags

    obj.__dict__.update(subst_dict)

def build_dir(name, subdir):
    if is_child():
        return os.path.join('build', name, subdir)
    else:
        return os.path.join('build', subdir)

# Doxygen API documentation
def build_dox(bld, name, version, srcdir, blddir, outdir=''):
    if not bld.env['DOCS']:
        return

    if is_child():
        src_dir = os.path.join(srcdir, name.lower())
        doc_dir = os.path.join(blddir, name.lower(), 'doc')
    else:
        src_dir = srcdir
        doc_dir = os.path.join(blddir, 'doc')

    subst_tg = bld(features     = 'subst',
                   source       = 'doc/reference.doxygen.in',
                   target       = 'doc/reference.doxygen',
                   install_path = '',
                   name         = 'doxyfile')

    subst_dict = {
        name + '_VERSION' : version,
        name + '_SRCDIR'  : os.path.abspath(src_dir),
        name + '_DOC_DIR' : os.path.abspath(doc_dir)
        }

    subst_tg.__dict__.update(subst_dict)

    subst_tg.post()

    docs = bld(features = 'doxygen',
               doxyfile = 'doc/reference.doxygen')

    docs.post()

    major = int(version[0:version.find('.')])
    bld.install_files(
        os.path.join('${DOCDIR}', '%s-%d' % (name.lower(), major), outdir, 'html'),
        bld.path.get_bld().ant_glob('doc/html/*'))
    for i in range(1, 8):
        bld.install_files('${MANDIR}/man%d' % i,
                          bld.path.get_bld().ant_glob('doc/man/man%d/*' % i,
                                                      excl='**/_*'))

# Version code file generation
def build_version_files(header_path, source_path, domain, major, minor, micro, exportname, visheader):
    header_path = os.path.abspath(header_path)
    source_path = os.path.abspath(source_path)
    text  = "int " + domain + "_major_version = " + str(major) + ";\n"
    text += "int " + domain + "_minor_version = " + str(minor) + ";\n"
    text += "int " + domain + "_micro_version = " + str(micro) + ";\n"
    try:
        o = open(source_path, 'w')
        o.write(text)
        o.close()
    except IOError:
        Logs.error('Failed to open %s for writing\n' % source_path)
        sys.exit(-1)

    text  = "#ifndef __" + domain + "_version_h__\n"
    text += "#define __" + domain + "_version_h__\n"
    if visheader != '':
        text += "#include \"" + visheader + "\"\n"
    text += exportname + " extern const char* " + domain + "_revision;\n"
    text += exportname + " extern int " + domain + "_major_version;\n"
    text += exportname + " extern int " + domain + "_minor_version;\n"
    text += exportname + " extern int " + domain + "_micro_version;\n"
    text += "#endif /* __" + domain + "_version_h__ */\n"
    try:
        o = open(header_path, 'w')
        o.write(text)
        o.close()
    except IOError:
        Logs.warn('Failed to open %s for writing\n' % header_path)
        sys.exit(-1)

    return None

# Internationalization with gettext
def build_i18n_pot(bld, srcdir, dir, name, sources, copyright_holder=None):
    Logs.info('Generating pot file from %s' % name)
    pot_file = '%s.pot' % name

    cmd = ['xgettext',
            '--keyword=_',
            '--keyword=N_',
            '--keyword=S_',
            '--keyword=P_:1,2',
            '--keyword=NC_:1c,2',
            '--from-code=UTF-8',
            '-o', pot_file]

    if copyright_holder:
        cmd += ['--copyright-holder="%s"' % copyright_holder]

    cmd += sources
    Logs.info('Updating ' + pot_file)
    subprocess.call(cmd, cwd=os.path.join(srcdir, dir))

def build_i18n_po(bld, srcdir, dir, name, sources, copyright_holder=None):
    pwd = os.getcwd()
    os.chdir(os.path.join(srcdir, dir))
    pot_file = '%s.pot' % name
    po_files = glob.glob('po/*.po')
    for po_file in po_files:
        cmd = ['msgmerge',
               '--update',
               '--no-fuzzy-matching',
               po_file,
               pot_file]
        Logs.info('Updating ' + po_file)
        subprocess.call(cmd)
    os.chdir(pwd)

def build_i18n_mo(bld, srcdir, dir, name, sources, copyright_holder=None):
    pwd = os.getcwd()
    os.chdir(os.path.join(srcdir, dir))
    pot_file = '%s.pot' % name
    po_files = glob.glob('po/*.po')
    for po_file in po_files:
        mo_file = po_file.replace('.po', '.mo')
        cmd = ['msgfmt',
               '-c',
               '-f',
               '-o',
               mo_file,
               po_file]
        Logs.info('Generating ' + po_file)
        subprocess.call(cmd)
    os.chdir(pwd)

def build_i18n(bld, srcdir, dir, name, sources, copyright_holder=None):
    build_i18n_pot(bld, srcdir, dir, name, sources, copyright_holder)
    build_i18n_po(bld, srcdir, dir, name, sources, copyright_holder)
    build_i18n_mo(bld, srcdir, dir, name, sources, copyright_holder)

def cd_to_build_dir(ctx, appname):
    orig_dir  = os.path.abspath(os.curdir)
    top_level = (len(ctx.stack_path) > 1)
    if top_level:
        os.chdir(os.path.join('build', appname))
    else:
        os.chdir('build')
    Logs.pprint('GREEN', "Waf: Entering directory `%s'" % os.path.abspath(os.getcwd()))

def cd_to_orig_dir(ctx, child):
    if child:
        os.chdir(os.path.join('..', '..'))
    else:
        os.chdir('..')

def pre_test(ctx, appname, dirs=['src']):
    diropts  = ''
    for i in dirs:
        diropts += ' -d ' + i
    cd_to_build_dir(ctx, appname)
    clear_log = open('lcov-clear.log', 'w')
    try:
        try:
            # Clear coverage data
            subprocess.call(('lcov %s -z' % diropts).split(),
                            stdout=clear_log, stderr=clear_log)
        except:
            Logs.warn('Failed to run lcov, no coverage report will be generated')
    finally:
        clear_log.close()

def post_test(ctx, appname, dirs=['src'], remove=['*boost*', 'c++*']):
    diropts  = ''
    for i in dirs:
        diropts += ' -d ' + i
    coverage_log           = open('lcov-coverage.log', 'w')
    coverage_lcov          = open('coverage.lcov', 'w')
    coverage_stripped_lcov = open('coverage-stripped.lcov', 'w')
    try:
        try:
            base = '.'
            if g_is_child:
                base = '..'

            # Generate coverage data
            subprocess.call(('lcov -c %s -b %s' % (diropts, base)).split(),
                            stdout=coverage_lcov, stderr=coverage_log)

            # Strip unwanted stuff
            subprocess.call(
                ['lcov', '--remove', 'coverage.lcov'] + remove,
                stdout=coverage_stripped_lcov, stderr=coverage_log)

            # Generate HTML coverage output
            if not os.path.isdir('coverage'):
                os.makedirs('coverage')
            subprocess.call('genhtml -o coverage coverage-stripped.lcov'.split(),
                            stdout=coverage_log, stderr=coverage_log)

        except:
            Logs.warn('Failed to run lcov, no coverage report will be generated')
    finally:
        coverage_stripped_lcov.close()
        coverage_lcov.close()
        coverage_log.close()

        print('')
        Logs.pprint('GREEN', "Waf: Leaving directory `%s'" % os.path.abspath(os.getcwd()))
        top_level = (len(ctx.stack_path) > 1)
        if top_level:
            cd_to_orig_dir(ctx, top_level)

    print('')
    Logs.pprint('BOLD', 'Coverage:', sep='')
    print('<file://%s>\n\n' % os.path.abspath('coverage/index.html'))

def run_tests(ctx, appname, tests, desired_status=0, dirs=['src'], name='*'):
    failures = 0
    diropts  = ''
    for i in dirs:
        diropts += ' -d ' + i

    # Run all tests
    for i in tests:
        s = i
        if type(i) == type([]):
            s = ' '.join(i)
        print('')
        Logs.pprint('BOLD', '** Test', sep='')
        Logs.pprint('NORMAL', '%s' % s)
        cmd = i
        if Options.options.grind:
            cmd = 'valgrind ' + i
        if subprocess.call(cmd, shell=True) == desired_status:
            Logs.pprint('GREEN', '** Pass')
        else:
            failures += 1
            Logs.pprint('RED', '** FAIL')

    print('')
    if failures == 0:
        Logs.pprint('GREEN', '** Pass: All %s.%s tests passed' % (appname, name))
    else:
        Logs.pprint('RED', '** FAIL: %d %s.%s tests failed' % (failures, appname, name))



#!/usr/bin/python -tt
#
# Copyright (C) 2005-2012 Erik de Castro Lopo <erikd@mega-nerd.com>
#
# Released under the 2 clause BSD license.

"""
This program checks C code for compliance to coding standards used in
libsndfile and other projects I run.
"""

import re
import sys


class Preprocessor:
	"""
	Preprocess lines of C code to make it easier for the CStyleChecker class to
	test for correctness. Preprocessing works on a single line at a time but
	maintains state between consecutive lines so it can preprocessess multi-line
	comments.
	Preprocessing involves:
	  - Strip C++ style comments from a line.
	  - Strip C comments from a series of lines. When a C comment starts and
	    ends on the same line it will be replaced with 'comment'.
	  - Replace arbitrary C strings with the zero length string.
	  - Replace '#define f(x)' with '#define f (c)' (The C #define requires that
	    there be no space between defined macro name and the open paren of the
	    argument list).
	Used by the CStyleChecker class.
	"""
	def __init__ (self):
		self.comment_nest = 0
		self.leading_space_re = re.compile ('^(\t+| )')
		self.trailing_space_re = re.compile ('(\t+| )$')
		self.define_hack_re = re.compile ("(#\s*define\s+[a-zA-Z0-9_]+)\(")

	def comment_nesting (self):
		"""
		Return the correct comment nesting. At the start and end of the file,
		this value should be zero. Inside C comments it should be 1 or
		(possibly) more.
		"""
		return self.comment_nest

	def __call__ (self, line):
		"""
		Strip the provided line of C and C++ comments. Stripping of multi-line
		C comments works as expected.
		"""

		line = self.define_hack_re.sub (r'\1 (', line)

		line = self.process_strings (line)

		# Strip C++ style comments.
		if self.comment_nest == 0:
			line = re.sub ("( |\t*)//.*", '', line)

		# Strip C style comments.
		open_comment = line.find ('/*')
		close_comment = line.find ('*/')

		if self.comment_nest > 0 and close_comment < 0:
			# Inside a comment block that does not close on this line.
			return ""

		if open_comment >= 0 and close_comment < 0:
			# A comment begins on this line but doesn't close on this line.
			self.comment_nest += 1
			return self.trailing_space_re.sub ('', line [:open_comment])

		if open_comment < 0 and close_comment >= 0:
			# Currently open comment ends on this line.
			self.comment_nest -= 1
			return self.trailing_space_re.sub ('', line [close_comment + 2:])

		if open_comment >= 0 and close_comment > 0 and self.comment_nest == 0:
			# Comment begins and ends on this line. Replace it with 'comment'
			# so we don't need to check whitespace before and after the comment
			# we're removing.
			newline = line [:open_comment] + "comment" + line [close_comment + 2:]
			return self.__call__ (newline)

		return line

	def process_strings (self, line):
		"""
		Given a line of C code, return a string where all literal C strings have
		been replaced with the empty string literal "".
		"""
		for k in range (0, len (line)):
			if line [k] == '"':
				start = k
				for k in range (start + 1, len (line)):
					if line [k] == '"' and line [k - 1] != '\\':
						return line [:start + 1] + '"' + self.process_strings (line [k + 1:])
		return line


class CStyleChecker:
	"""
	A class for checking the whitespace and layout of a C code.
	"""
	def __init__ (self, debug):
		self.debug = debug
		self.filename = None
		self.error_count = 0
		self.line_num = 1
		self.orig_line = ''
		self.trailing_newline_re = re.compile ('[\r\n]+$')
		self.indent_re = re.compile ("^\s*")
		self.last_line_indent = ""
		self.last_line_indent_curly = False
		self.error_checks = \
                        [ ( re.compile ("^ "),          "leading space as indentation instead of tab - use tabs to indent, spaces to align" )
                        ]                                                                                                                                       
		self.warning_checks = \
                        [   ( re.compile ("{[^\s]"),      "missing space after open brace" )
                          , ( re.compile ("[^\s]}"),      "missing space before close brace" )
                          , ( re.compile ("^[ \t]+$"),     "empty line contains whitespace" )
                          , ( re.compile ("[^\s][ \t]+$"),     "contains trailing whitespace" )

                          , ( re.compile (",[^\s\n]"),            "missing space after comma" )
                          , ( re.compile (";[a-zA-Z0-9]"),        "missing space after semi-colon" )
                          , ( re.compile ("=[^\s\"'=]"),          "missing space after assignment" )
                          
                          # Open and close parenthesis.
                          , ( re.compile ("[^_\s\(\[\*&']\("),             "missing space before open parenthesis" )
                          , ( re.compile ("\)(-[^>]|[^;,'\s\n\)\]-])"),    "missing space after close parenthesis" )
                          , ( re.compile ("\( [^;]"),                     "space after open parenthesis" )
                          , ( re.compile ("[^;] \)"),                     "space before close parenthesis" )

                          # Open and close square brace.
                          , ( re.compile ("\[ "),                                 "space after open square brace" )
                          , ( re.compile (" \]"),                                 "space before close square brace" )

                          # Space around operators.
                          , ( re.compile ("[^\s][\*/%+-][=][^\s]"),               "missing space around opassign" )
                          , ( re.compile ("[^\s][<>!=^/][=]{1,2}[^\s]"),  "missing space around comparison" )

                          # Parens around single argument to return.
                          , ( re.compile ("\s+return\s+\([a-zA-Z0-9_]+\)\s+;"),   "parens around return value" )
                        ]                                                                                                                                       

                

	def get_error_count (self):
		"""
		Return the current error count for this CStyleChecker object.
		"""
		return self.error_count

	def check_files (self, files):
		"""
		Run the style checker on all the specified files.
		"""
		for filename in files:
			self.check_file (filename)

	def check_file (self, filename):
		"""
		Run the style checker on the specified file.
		"""
		self.filename = filename
		try:
			cfile = open (filename, "r")
		except IOError as e:
			return

		self.line_num = 1

		preprocess = Preprocessor ()
		while 1:
			line = cfile.readline ()
			if not line:
				break

			line = self.trailing_newline_re.sub ('', line)
			self.orig_line = line

			self.line_checks (preprocess (line))

			self.line_num += 1

		cfile.close ()
		self.filename = None

		# Check for errors finding comments.
		if preprocess.comment_nesting () != 0:
			print ("Weird, comments nested incorrectly.")
			sys.exit (1)

		return

	def line_checks (self, line):
		"""
		Run the style checker on provided line of text, but within the context
		of how the line fits within the file.
		"""

		indent = len (self.indent_re.search (line).group ())
		if re.search ("^\s+}", line):
			if not self.last_line_indent_curly and indent != self.last_line_indent:
				None	# self.error ("bad indent on close curly brace")
			self.last_line_indent_curly = True
		else:
			self.last_line_indent_curly = False

		# Now all the stylistic warnings regex checks.
		for (check_re, msg) in self.warning_checks:
			if check_re.search (line):
				self.warning (msg)

                # Now all the stylistic error regex checks.
		for (check_re, msg) in self.error_checks:
			if check_re.search (line):
				self.error (msg)

                                
		if re.search ("[a-zA-Z0-9_][<>!=^/&\|]{1,2}[a-zA-Z0-9_]", line):
                        # ignore #include <foo.h> and C++ templates with indirection/pointer/reference operators
			if not re.search (".*#include.*[a-zA-Z0-9]/[a-zA-Z]", line) and not re.search ("[a-zA-Z0-9_]>[&\*]*\s", line):
				self.error ("missing space around operator")

		self.last_line_indent = indent
		return

	def error (self, msg):
		"""
		Print an error message and increment the error count.
		"""
		print ("%s (%d) : STYLE ERROR %s" % (self.filename, self.line_num, msg))
		if self.debug:
			print ("'" + self.orig_line + "'")
		self.error_count += 1

	def warning (self, msg):
		"""
		Print a warning message and increment the error count.
		"""
		print ("%s (%d) : STYLE WARNING %s" % (self.filename, self.line_num, msg))
		if self.debug:
			print ("'" + self.orig_line + "'")
                
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

if len (sys.argv) < 1:
	print ("Usage : yada yada")
	sys.exit (1)

# Create a new CStyleChecker object
if sys.argv [1] == '-d' or sys.argv [1] == '--debug':
	cstyle = CStyleChecker (True)
	cstyle.check_files (sys.argv [2:])
else:
	cstyle = CStyleChecker (False)
	cstyle.check_files (sys.argv [1:])


if cstyle.get_error_count ():
	sys.exit (1)

sys.exit (0)



#! /usr/bin/python

# Session Exchange
# By Taybin Rutkin
# Copyright 2004-2005, under the GPL

VERSION='0.1.2'

#twisted libraries
from twisted.internet import gtk2reactor
gtk2reactor.install()
from twisted.internet import reactor, protocol
import twisted.internet.error

#pygtk libraries
import gobject
import gtk

#standard python2.2 libraries
import getopt
import os
import os.path
import re
import shelve
import string
import sys
import xml.dom.minidom

v2paths = True

def get_header_size(filename):
        size = 0
        file = open(filename, 'r')
        while True:
                chunk = file.read(4)
                size += 4
                if chunk == "data":
                        file.close()
                        return size + 4        #include the size chunk after "data"
                if not chunk:
                        file.close()
                        return None

def append_empty_data(self, filename, size):
        file = open(filename, 'a')
        file.seek(size-1)
        file.write('\x00')
        file.close()
        
def get_sound_list(snapshot):
        doc = xml.dom.minidom.parse(snapshot)
        
        regionlist = []
        playlists_tag = doc.getElementsByTagName('Playlists')
        playlists = playlists_tag[0].getElementsByTagName('Playlist')
        for play in playlists:
                regions = play.getElementsByTagName('Region')
                for region in regions:
                        regionlist.append(region.getAttribute('source-0'))
                        regionlist.append(region.getAttribute('source-1'))
                        regionlist.append(region.getAttribute('source-2'))
                        regionlist.append(region.getAttribute('source-3'))
                        regionlist.append(region.getAttribute('source-4'))
                        regionlist.append(region.getAttribute('source-5'))
        
        sourcelist = {}
        sources = doc.getElementsByTagName('Source')
        for source in sources:
                sourcelist[source.getAttribute('id')] = str(source.getAttribute('name'))

        soundlist = []
        for id in regionlist:
                if sourcelist.has_key(id):
                        soundlist.append(sourcelist[id])
        
        return soundlist

def raise_error(string, parent):
        dialog = gtk.MessageDialog(parent, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
        gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, string)
        
        dialog.run()
        dialog.destroy()

class Data(object):
        def delete_snap(self, session, collab, snap):
                sessions = self._data['sessions']
                sessions[session]['collabs'][collab]['snaps'].remove(snap)
                self._data['sessions'] = sessions
        
        def delete_collab(self,session, collab):
                sessions = self._data['sessions']
                del sessions[session]['collabs'][collab]
                self._data['sessions'] = sessions
        
        def delete_session(self, session):
                sessions = self._data['sessions']
                del sessions[session]
                self._data['sessions'] = sessions
        
        def add_snap(self, session_name, collab_name, snap_name):
                sessions = self._data['sessions']
                sessions[session_name]['collabs'][collab_name]['snaps'].append(snap_name)
                sessions[session_name]['collabs'][collab_name]['snaps'].sort()
                self._data['sessions'] = sessions
                
                g_display.update_snap_view()
        
        def add_collab(self, session_name, collab_name, ip_address, port):
                sessions = self._data['sessions']
                sessions[session_name]['collabs'][collab_name] = {}
                sessions[session_name]['collabs'][collab_name]['snaps'] = []
                sessions[session_name]['collabs'][collab_name]['sounds'] = []
                sessions[session_name]['collabs'][collab_name]['ip'] = ip_address
                sessions[session_name]['collabs'][collab_name]['port'] = port
                sessions[session_name]['collabs'][collab_name]['v2paths'] = True
                self._data['sessions'] = sessions
                
                client = ExchangeClientFactory(session_name, collab_name, None, self.debug_mode)
                reactor.connectTCP(ip_address, port, client)
                g_display.show_status("connecting")
                
                g_display.update_collab_view()
        
        def add_session(self, session_path):
                sessions = self._data['sessions']
                
                session_name = session_path[session_path.rfind('/', 0, len(session_path)-2)+1: -1]
                sessions[session_name] = {}
                sessions[session_name]['path'] = session_path 
                sessions[session_name]['collabs'] = {}
                sessions[session_name]['collabs'][self._data['user']] = {}
                sessions[session_name]['collabs'][self._data['user']]['snaps'] = []
                sessions[session_name]['collabs'][self._data['user']]['sounds'] = []
                if os.path.isdir (os.path.join (session_path,'sounds')):
                        sessions[session_name]['collabs'][self._data['user']]['v2paths'] = False
                        v2paths = False
                else:
                        sessions[session_name]['collabs'][self._data['user']]['v2paths'] = True
                
                self._data['sessions'] = sessions
                
                self.rescan_session(session_name)

        def rescan_session(self, session_name):
                sessions = self._data['sessions']
                
                session_path = sessions[session_name]['path']
                sessions[session_name]['collabs'][self._data['user']]['snaps'] = self._scan_snapshots(session_path)
                sessions[session_name]['collabs'][self._data['user']]['sounds'] = self._scan_sounds(session_name)
                
                self._data['sessions'] = sessions
                
                g_display.update_snap_view()
                
                print self._data['sessions']
        
        def create_session(self, session_path):
                sessions = self._data['sessions']

                session_name = session_path[session_path.rfind('/', 0, len(session_path)-2)+1: ]
                try:
                        os.mkdir(session_path)
                        os.mkdir(os.path.join (session_path,'interchange'))
                        os.mkdir(os.path.join (session_path,'interchange',session_name))
                        os.mkdir(os.path.join (session_path,'interchange',session_name,'audiofiles'))
                except OSError:
                        raise_error("Could not create session directory", g_display.window)
                        return
                
                sessions[session_name] = {}
                sessions[session_name]['path'] = session_path
                sessions[session_name]['collabs'] = {}
                sessions[session_name]['collabs'][self._data['user']] = {}
                sessions[session_name]['collabs'][self._data['user']]['snaps'] = []
                sessions[session_name]['collabs'][self._data['user']]['sounds'] = []
                
                self._data['sessions'] = sessions
                print self._data['sessions']
        
        def get_session_path(self, session):
                sessions = self._data['sessions']
                return sessions[session]['path']
        
        def get_user(self):
                return self._data['user']
        
        def set_user(self, username):
                self._data['user'] = username
        
        def get_collab_ip(self, session, collab):
                sessions = self._data['sessions']
                return sessions[session]['collabs'][collab]['ip']
        
        def close(self):
                self._data.close()
        
        def get_sessions(self):
                sessions = self._data['sessions']
                sess = sessions.keys()
                sess.sort()
                return sess
        
        def get_collabs(self, session):
                if session:
                        sessions = self._data['sessions']
                        collabs = sessions[session]['collabs'].keys()
                        collabs.sort()
                        return collabs
                else:
                        return []
        
        def get_snaps(self, session, collab):
                if session and collab:
                        sessions = self._data['sessions']
                        snaps = sessions[session]['collabs'][collab]['snaps']
                        snaps.sort()
                        return snaps
                else:
                        return []
        
        def get_sounds(self, session, collab):
                if session and collab:
                        sessions = self._data['sessions']
                        sounds = sessions[session]['collabs'][self._data['user']]['sounds']
                        sounds.sort()
                        return sounds
                else:
                        return []
                
        def _scan_snapshots(self, session):
                snaps = []
                files = os.listdir(session)
                pattern = re.compile(r'\.ardour$')
                for file in files:
                        if pattern.search(file):
                                snaps.append(file[0:-7])
                                print file[0:-7]
                return snaps
        
        def _scan_sounds(self, session):
                sessions = self._data['sessions']

                sounds = []
                if v2paths:
                        print session
                        print os.path.join (sessions[session]['path'],'interchange', session, 'audiofiles')
                        files = os.listdir(os.path.join (sessions[session]['path'],'interchange', session, 'audiofiles'))
                else:
                        files = os.listdir(os.path.join (session,'sounds'))
                pattern = re.compile(r'\.peak$')
                for file in files:
                        if not pattern.search(file):
                                sounds.append(file)
                return sounds
        
        def __init__(self, *args):
                self._data = shelve.open(os.path.expanduser('~/.session_exchange'), 'c')
                self.port = 8970
                self.debug_mode = False
                if len(self._data.keys()) < 1:
                        self._data['sessions'] = {}
                        self._data['user'] = ''
                
                self._collabs = {}

from twisted.protocols.basic import FileSender
class FileSenderLimited(FileSender):
        def beginFileTransfer(self, file, consumer, limit, transform = None):
                self.file = file
                self.consumer = consumer
                self.CHUNK_SIZE = limit
                self.transform = transform
                
                self.consumer.registerProducer(self, False)
                self.deferred = defer.Deferred()
                return self.deferred
        
        def resumeProducing(self):
                chunk = ''
                chunk = self.file.read(self.CHUNK_SIZE)
                
                if self.transform:
                        chunk = self.transform(chunk)

                self.consumer.write(chunk)
                self.lastSent = chunk[-1]
                self.file = None
                self.consumer.unregisterProducer()
                self.deferred.callback(self.lastSent)
                self.deferred = None

from twisted.protocols.basic import LineReceiver
class ExchangeServer (LineReceiver):
        def __init__(self):
                self.state = "IDLE"
        
        def error(self, message):
                self.sendLine("ERROR")
                self.sendLine(message)
                self.transport.loseConnection()
        
        def connectionLost(self, reason):
                print "server: connection lost: ", reason
        
        def connectionMade(self):
                print "server: connection made"
        
        def lineReceived(self, data):
                print "server: ", data
                
                if self.state == "SESSION":
                        if g_data.get_sessions().count(data):
                                self.session_name = data
                                self.state = "IDLE"
                                self.sendLine("OK")
                        else:
                                self.error(data + " doesn't exist on server")
                elif self.state == "SNAPSHOT":
                        if g_data.get_snaps(self.session_name, g_data.get_user()).count(data):
                                filename = g_data.get_session_path(self.session_name)+data+'.ardour'
                                print filename
                                self.sendLine(str(os.stat(filename).st_size))
                                self.sendLine("OK")
                                self.file = open(filename, 'r')
                                file_sender = FileSender()
                                cb = file_sender.beginFileTransfer(self.file, self.transport)
                                cb.addCallback(self.file_done)
                        else:
                                self.error("snapshot: " + data + " doesn't exist on server")
                elif self.state == "SOUNDFILE" or self.state == "SOUNDFILE_HEADER":
                        if g_data.get_sounds(self.session_name, g_data.get_user()).count(data):
                                filename = g_data.get_session_path(self.session_name)+"/interchange/"+self.session_name+"/audiofiles/"+data
                                print filename
                                if self.state == "SOUNDFILE":
                                        self.sendLine(str(os.stat(filename).st_size))
                                else:        #SOUNDFILE_HEADER
                                        header_size = get_header_size(filename)
                                        if header_size:
                                                self.sendLine(str(header_size))
                                        else:
                                                self.error('soundfile: ' + data + 'doesn\'t have "data" chunk')
                                self.sendLine("OK")
                                self.file = open(filename, 'r')
                                if self.state == "SOUNDFILE":
                                        file_sender = FileSender()
                                        cb = file_sender.beginFileTransfer(self.file, self.transport)
                                else:        # SOUNDFILE_HEADER
                                        file_sender = FileSenderLimited()
                                        cb = file_sender.beginFileTransfer(self.file, self.transport, header_size)
                                cb.addCallback(self.file_done)
                        else:
                                self.error("soundfile: " + data + "doesn't exist on server")
                elif self.state == "SOUNDFILE_SIZE":
                        if g_data.get_sounds(self.session_name, g_data.get_user()).count(data):
                                filename = g_data.get_session_path(self.session_name)+"/sounds/"+data
                                print filename
                                self.sendLine(str(os.stat(filename).st_size))
                                self.state = "IDLE"
                elif data == "SESSION":
                        self.state = "SESSION"
                elif data == "SNAPS":
                        self.state = "SNAPS"
                        for snap in g_data.get_snaps(self.session_name, g_data.get_user()):
                                self.sendLine(snap)
                        self.sendLine("OK")
                        self.state = "IDLE"
                elif data == "SNAPSHOT":
                        self.state = "SNAPSHOT"
                elif data == "SOUNDFILE":
                        self.state = "SOUNDFILE"
                elif data == "SOUNDFILE_HEADER":
                        self.state = "SOUNDFILE_HEADER"
                elif data == "SOUNDFILE_SIZE":
                        self.state = "SOUNDFILE_SIZE"
        
        def file_done(self, data):
                print "server: file done"
                self.file.close()
                self.state = "IDLE"
        
class ExchangeServerFactory(protocol.ServerFactory):
        protocol = ExchangeServer
        
        def __init__(self):
                pass

class ExchangeClient (LineReceiver):
        def __init__(self, session_name, collab_name, snap_name, debug_mode):
                self.session_name = session_name
                self.collab_name = collab_name
                self.snap_name = snap_name
                self.debug_mode = debug_mode
                self.state = "IDLE"
        
        def connectionLost(self, reason):
                g_display.show_status("Connection lost")
        
        def connectionMade(self):
                g_display.show_status("Connection made")
                self.state = "SESSION"
                self.sendLine("SESSION")
                self.sendLine(self.session_name)
        
        def rawDataReceived(self, data):
                self.file.write(data)
                self.received += len(data)
                print self.received, self.filesize
                if self.received >= self.filesize:
                        self.setLineMode()
                        self.file.close()
                        g_data.rescan_session(self.session_name)
                        if self.state == "SNAPSHOT":
                                self.sounds = get_sound_list(self.filename)
                                if len(self.sounds):
                                        self.sound_index = 0
                                        if self.debug_mode:
                                                self.state = "SOUNDFILE_HEADER"
                                                self.sendLine("SOUNDFILE_HEADER")
                                        else:
                                                self.state = "SOUNDFILE"
                                                self.sendLine("SOUNDFILE")
                                        self.sendLine(self.sounds[self.sound_index])
                                else:
                                        self.transport.loseConnection()
                        elif self.state == "SOUNDFILE":
                                self.sound_index += 1
                                if self.sound_index > len(self.sounds)-1:
                                        self.transport.loseConnection()
                                else:
                                        self.sendLine("SOUNDFILE")
                                        self.sendLine(self.sounds[self.sound_index])
                        elif self.state == "SOUNDFILE_HEADER":
                                self.state = "SOUNDFILE_SIZE"
                                self.sendLine("SOUNDFILE_SIZE")
                                self.sendLine(self.sounds[self.sound_index])
        
        def lineReceived(self, data):
                print "client: ", data
                
                if data == "ERROR":
                        self.state = "ERROR"
                elif data == "OK":
                        if self.state == "SESSION":
                                if self.snap_name:
                                        self.state = "SNAPSHOT"
                                        self.sendLine("SNAPSHOT")
                                        self.sendLine(self.snap_name)
                                else:
                                        self.state = "SNAPS"
                                        self.sendLine("SNAPS")
                        elif self.state == "SNAPS":
                                self.transport.loseConnection()
                        elif self.state == "SNAPSHOT":
                                self.setRawMode()
                                self.filename = g_data.get_session_path(self.session_name)+'/'+self.snap_name+'.ardour'
                                self.file = open(self.filename, 'w')
                                self.received = 0
                        elif self.state == "SOUNDFILE" or self.state == "SOUNDFILE_HEADER":
                                self.setRawMode()
                                self.filename = g_data.get_session_path(self.session_name)+"/interchange/"+self.session_name+"/audiofiles/"+self.sounds[self.sound_index]
                                self.file = open(self.filename, 'w')
                                self.received = 0
                elif self.state == "ERROR":
                        raise_error(data, g_display.window)
                elif self.state == "SNAPS":
                        g_data.add_snap(self.session_name, self.collab_name, data)
                elif self.state == "SNAPSHOT":
                        self.filesize = int(data)
                elif self.state == "SOUNDFILE":
                        self.filesize = int(data)
                elif self.state == "SOUNDFILE_HEADER":
                        self.filesize = int(data)
                elif self.state == "SOUNDFILE_SIZE":
                        append_empty_data(self.filename, int(data))
                        self.sound_index += 1
                        if self.sound_index > len(self.sounds)-1:
                                self.transport.loseConnection()
                        else:
                                self.state = "SOUNDFILE_HEADER"
                                self.sendLine("SOUNDFILE_HEADER")
                                self.sendLine(self.sounds[self.sound_index])

class ExchangeClientFactory(protocol.ClientFactory):
        def buildProtocol(self, addr):
                return ExchangeClient(self.session_name, self.collab_name, self.snap_name, self.debug_mode)
        
        def clientConnectionFailed(self, connector, reason):
                raise_error('Connection failed: ' + reason.getErrorMessage(), g_display.window)
                g_display.show_status('Connection failed')
        
        def __init__(self, session_name, collab_name, snap_name, debug_mode):
                self.session_name = session_name
                self.collab_name = collab_name
                self.snap_name = snap_name
                self.debug_mode = debug_mode

class HelperWin(object):
        def delete_me(self, window):
                self = 0

class Preferences(HelperWin):
        def __init__(self):
                self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
                self.window.set_title('Preferences')
                self.window.connect('destroy', self.delete_me)
                self.window.set_position(gtk.WIN_POS_MOUSE)
                
                main_box = gtk.VBox()
                self.window.add(main_box)
                
                hbox1 = gtk.HBox()
                label1 = gtk.Label("User")
                self.user = gtk.Entry()
                self.user.set_text(g_data.get_user())
                hbox1.pack_start(label1)
                hbox1.pack_start(self.user)
                main_box.pack_start(hbox1)
                
                ok_btn = gtk.Button("Ok")
                ok_btn.connect('clicked', self.ok_clicked)
                main_box.pack_start(ok_btn)
                
                self.window.show_all()
                
        def ok_clicked(self, btn):
                g_data.set_user(self.user.get_text())
                self.window.hide_all()
                
        def show_all(self):
                self.window.show_all()

class AddCollaborator(HelperWin):
        def __init__(self, session):
                self.session_name = session
                
                self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
                self.window.set_title('Fetch Session')
                self.window.connect('destroy', self.delete_me)
                self.window.set_position(gtk.WIN_POS_MOUSE)
                
                main_box = gtk.VBox()
                self.window.add(main_box)
                
                hbox0 = gtk.HBox()
                label0 = gtk.Label("Collaborator")
                self.collab = gtk.Entry()
                self.collab.connect('key-release-event', self.key_press)
                hbox0.pack_start(label0)
                hbox0.pack_start(self.collab)
                main_box.pack_start(hbox0)
                
                hbox1 = gtk.HBox()
                label1 = gtk.Label("IP Address")
                self.address = gtk.Entry()
                self.address.connect('key-release-event', self.key_press)
                hbox1.pack_start(label1)
                hbox1.pack_start(self.address)
                main_box.pack_start(hbox1)
                
                hbox2 = gtk.HBox()
                label2 = gtk.Label("Port Number")
                self.port = gtk.Entry()
                self.port.connect('key-release-event', self.key_press)
                self.port.set_text(str(g_data.port))
                hbox2.pack_start(label2)
                hbox2.pack_start(self.port)
                main_box.pack_start(hbox2)
                
                hbox3 = gtk.HBox()
                label3 = gtk.Label("Username")
                label3.set_sensitive(False)
                self.username = gtk.Entry()
                self.username.set_sensitive(False)
                hbox3.pack_start(label3)
                hbox3.pack_start(self.username)
                main_box.pack_start(hbox3)
                
                hbox4 = gtk.HBox()
                label4 = gtk.Label("Password")
                label4.set_sensitive(False)
                self.password = gtk.Entry()
                self.password.set_sensitive(False)
                hbox4.pack_start(label4)
                hbox4.pack_start(self.password)
                main_box.pack_start(hbox4)
                
                self.ok_btn = gtk.Button(gtk.STOCK_OK)
                self.ok_btn.set_use_stock(True)
                self.ok_btn.connect('clicked', self.ok_clicked)
                self.ok_btn.set_sensitive(False)
                main_box.pack_start(self.ok_btn)
                
                self.window.show_all()
        
        def key_press(self, event, data):
                if self.collab.get_text() and self.address.get_text() and self.port.get_text():
                        self.ok_btn.set_sensitive(True)
                else:
                        self.ok_btn.set_sensitive(False)
                return True
        
        def ok_clicked(self, btn):
                self.window.hide_all()
                g_data.add_collab(self.session_name, self.collab.get_text(), self.address.get_text(), int(self.port.get_text()))
                self.collab.set_text('')
                self.address.set_text('')
                self.port.set_text('')
                self.username.set_text('')
                self.password.set_text('')
        
        def show_all(self):
                self.window.show_all()

class ArdourShareWindow(object):
        def menuitem_cb(self, window, action, widget):
                print self, window, action, widget
        
        def add_collaborator_cb(self, window, action, widget):
                if self.session:
                        self.add_session = AddCollaborator(self.session)
        
        def fetch_snapshot_cb(self, window, action, widget):
                if self.session and self.collab and self.collab != g_data.get_user():
                        client = ExchangeClientFactory(self.session, self.collab, self.snap, g_data.debug_mode)
                        reactor.connectTCP(g_data.get_collab_ip(self.session, self.collab), g_data.port, client)
        
        def preferences_cb(self, window, action, widget):
                self.preferences = Preferences()
        
        def add_session_ok_file_btn_clicked(self, w):
                filename = self.file_sel.get_filename()
                if filename.endswith(".ardour"):
                        g_data.add_session(filename[0:filename.rfind("/")+1])
                        self.update_session_view()
                else:
                        raise_error("Not an Ardour session", self.window)
                self.file_sel.destroy()
        
        def add_session_cb(self, window, action, widget):
                if g_data.get_user():
                        self.file_sel = gtk.FileSelection("Add Session...")
                        self.file_sel.ok_button.connect("clicked", self.add_session_ok_file_btn_clicked)
                        self.file_sel.cancel_button.connect("clicked", lambda w: self.file_sel.destroy())
                        self.file_sel.connect("destroy", lambda w: self.file_sel.destroy())
                        self.file_sel.show()
                else:
                        raise_error("Set the user name in the preferences first", self.window)
        
        def create_session_cb(self, window, action, widget):
                if g_data.get_user():
                        self.file_sel = gtk.FileSelection("Create Session...")
                        self.file_sel.ok_button.connect("clicked", self.create_file_ok_btn_clicked)
                        self.file_sel.cancel_button.connect("clicked", lambda w: self.file_sel.destroy())
                        self.file_sel.connect("destroy", lambda w: self.file_sel.destroy())
                        self.file_sel.show()
                else:
                        raise_error("Set the user name in the preferences first", self.window)
        
        def create_file_ok_btn_clicked(self, w):
                filename = self.file_sel.get_filename()
                if len(filename) > 0:
                        g_data.create_session(filename)
                        self.update_session_view()
                else:
                        raise_error("Not an Ardour session", self.window)
                self.file_sel.destroy()
        
        def update_session_view(self):
                self.session_model.clear()
                for session in g_data.get_sessions():
                        self.session_model.set(self.session_model.append(), 0, session)
        
        def update_collab_view(self):
                self.collab_model.clear()
                for collab in g_data.get_collabs(self.session):
                        self.collab_model.set(self.collab_model.append(), 0, collab)
        
        def update_snap_view(self):
                self.snap_model.clear()
                for snap in g_data.get_snaps(self.session, self.collab):
                        self.snap_model.set(self.snap_model.append(), 0, snap)
        
        def cb_session_selection_changed(self, selection_object):
                selected = []
                selection_object.selected_foreach(lambda model, path, iter, sel = selected: sel.append(path))
                for x in selected:
                        self.session = self.session_model[x][0]
                self.selected_type = "session"
                self.update_collab_view()
        
        def cb_collab_selection_changed(self, selection_object):
                selected = []
                selection_object.selected_foreach(lambda model, path, iter, sel = selected: sel.append(path))
                for x in selected:
                        self.collab = self.collab_model[x][0]
                self.selected_type = "collab"
                self.update_snap_view()
        
        def cb_snap_selection_changed(self, selection_object):
                selected = []
                selection_object.selected_foreach(lambda model, path, iter, sel = selected: sel.append(path))
                for x in selected:
                        self.snap = self.snap_model[x][0]
                self.selected_type = "snap"
        
        def delete_cb(self, window, action, widget):
                if self.selected_type == "session":
                        g_data.delete_session(self.session)
                        self.session = ""
                        self.collab = ""
                        self.snap = ""
                elif self.selected_type == "collab":
                        g_data.delete_collab(self.session, self.collab)
                        self.collab = ""
                        self.snap = ""
                elif self.selected_type == "snap":
                        g_data.delete_snap(self.session, self.collab, self.snap)
                        self.snap = ""
                
                self.update_session_view()
                self.update_collab_view()
                self.update_snap_view()
                self.selected_type = ""
                
        def show_status(self, text):
                mid = self.status_bar.push(self._status_cid, text)
                if self._status_mid:
                        self.status_bar.remove(self._status_cid, self._status_mid)
                self._status_mid = mid
        
        def __init__(self):
                self.selected_type = ""
                self.session = ""
                self.collab = g_data.get_user()
                self.snap = ""
                
                self.preferences = 0
                self.add_collab = 0
                self.add_session = 0
                
                self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
                self.window.set_title('Session Exchange')
                self.window.set_size_request(400, 200)
                self.window.connect('destroy', lambda win: gtk.main_quit())
                self.window.set_position(gtk.WIN_POS_MOUSE)
                
                accel_group = gtk.AccelGroup()
                self.window.add_accel_group(accel_group)
                
                main_box = gtk.VBox()
                self.window.add(main_box)
                
                menu_items = (
                        ('/_File',            None,         None,             0, '<Branch>'),
                        ('/File/_Add Session...','<control>A', self.add_session_cb, 0, ''),
                        ('/File/Create _Session...', '<control>S', self.create_session_cb, 0, ''),
                        ('/File/sep1',        None,         None,             0, '<Separator>'),
                        ('/File/_Quit',       '<control>Q', gtk.main_quit,     0, '<StockItem>', gtk.STOCK_QUIT),
                        ('/_Edit',            None,         None,             0, '<Branch>' ),
                        ('/Edit/Cu_t',        '<control>X', self.menuitem_cb, 0, '<StockItem>', gtk.STOCK_CUT),
                        ('/Edit/_Copy',       '<control>C', self.menuitem_cb, 0, '<StockItem>', gtk.STOCK_COPY),
                        ('/Edit/_Paste',      '<control>V', self.menuitem_cb, 0, '<StockItem>', gtk.STOCK_PASTE),
                        ('/Edit/_Delete',     None,         self.delete_cb, 0, '<StockItem>', gtk.STOCK_DELETE),
                        ('/Edit/sep1',        None,         None,             0, '<Separator>'),
                        ('/Edit/Add Colla_borator...','<control>B', self.add_collaborator_cb,0,''),
                        ('/Edit/_Fetch Snapshot','<control>F', self.fetch_snapshot_cb,0,''),
                        ('/Edit/sep1',        None,         None,             0, '<Separator>'),
                        ('/Edit/_Preferences...','<control>P', self.preferences_cb, 0, '')
                )
                
                #need to hold a reference to the item_factory or the menubar will disappear.
                self.item_factory = gtk.ItemFactory(gtk.MenuBar, '<main>', accel_group)
                self.item_factory.create_items(menu_items, self.window)
                main_box.pack_start(self.item_factory.get_widget('<main>'), False)
                
                pane1 = gtk.HPaned()
                pane2 = gtk.HPaned()
                pane1.pack2(pane2, True, False)
                
                scroll1 = gtk.ScrolledWindow()
                scroll1.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
                pane1.pack1(scroll1, True, False)
                scroll2 = gtk.ScrolledWindow()
                scroll2.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
                pane2.pack1(scroll2, True, False)
                scroll3 = gtk.ScrolledWindow()
                scroll3.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
                pane2.pack2(scroll3, True, False)
                
                self.session_model = gtk.ListStore(gobject.TYPE_STRING)
                view1 = gtk.TreeView(self.session_model)
                column1 = gtk.TreeViewColumn('Sessions', gtk.CellRendererText(), text=0)
                view1.append_column(column1)
                self.session_selection = view1.get_selection()
                self.session_selection.connect("changed", self.cb_session_selection_changed)
                scroll1.add(view1)
                
                self.update_session_view()
                
                self.collab_model = gtk.ListStore(gobject.TYPE_STRING)
                view2 = gtk.TreeView(self.collab_model)
                column2 = gtk.TreeViewColumn('Collaborators', gtk.CellRendererText(), text=0)
                view2.append_column(column2)
                self.collab_selection = view2.get_selection()
                self.collab_selection.connect("changed", self.cb_collab_selection_changed)
                scroll2.add(view2)
                
                self.snap_model = gtk.ListStore(gobject.TYPE_STRING)
                view3 = gtk.TreeView(self.snap_model)
                column3 = gtk.TreeViewColumn('Snapshots', gtk.CellRendererText(), text=0)
                view3.append_column(column3)
                self.snap_selection = view3.get_selection()
                self.snap_selection.connect("changed", self.cb_snap_selection_changed)
                scroll3.add(view3)
                
                main_box.pack_start(pane1, True, True)
                
                self.status_bar = gtk.Statusbar()
                main_box.pack_start(self.status_bar, False)
                self._status_cid = self.status_bar.get_context_id('display')
                self._status_mid = ''
                
                self.window.show_all()

def print_help():
        print """
        -h, --help
        -n, --no-server          Only act as a client
        -p, --port <port number> Defaults to 8970
        -d, --debug              Infers audio files.  For debugging Ardour.
        -v, --version            Version
        """
        sys.exit(2)

def main():
        try:
                opts, args = getopt.getopt(sys.argv[1:], "hp:ndv", ["help", "port=", "no-server", "debug", "version"])
        except getopt.GetoptError:
                print_help()
        
        server = True
        for o, a in opts:
                if o in ("-h", "--help"):
                        print_help()
                if o in ("-d", "--debug"):
                        g_display.window.set_title('Session Exchange: Debug Mode')
                        g_data.debug_mode = True
                if o in ("-p", "--port"):
                        g_data.port = int(a)
                if o in ("-n", "--no-server"):
                        server = False
                if o in ("-v", "--version"):
                        print VERSION
                        sys.exit(2)
        
        if (server):
                try:
                        reactor.listenTCP(g_data.port, ExchangeServerFactory())
                except twisted.internet.error.CannotListenError:
                        print "Can not listen on a port number under 1024 unless run as root"
                        sys.exit(2)

        reactor.run()

        g_data.close()

# global objects
g_data = Data()
g_display = ArdourShareWindow()

if __name__ == '__main__':
        main()



import re
import subprocess
from waflib.Task import Task
from waflib.TaskGen import feature, after_method

class nm_def(Task):
    p = re.compile(r'^\s*([0-9a-fA-F]+)\s+([A-Za-z])\s+(\S+)')  # Parse llvm-nm output (grouped as address, symbol-type letter, symbol-name)
    s = re.compile(r'(?!__?imp_)[^?]*$|\?(?!\?_(?:[GE]|C@_))')  # Filters C/C++ symbols
    def __str__(self): return self.outputs[0].name              # Change what's said in the build log
    def run(self):
        p, s, nm = self.p.match, self.s.match, self.env.llvm_nm
        internal = re.compile(r'__(?:real|x86\.get_pc|GLOBAL__|tls_|imp_|guard_|fstack|asan_|msan_|eh|unwind|purecall|CRT|vcrt|MSVC)')
        self.outputs[0].write('EXPORTS\n' + '\n'.join(
            r
            for f in self.inputs                                        # For .o files
            for l in subprocess.check_output(nm + ['--defined-only', f.abspath()], text=1).splitlines() #Output of the llvm command run on the file
            if (m := p(l))                                              # Skip anything that isn't a symbol
            and m[2].isupper()                                          # Uppercase letter means globally visible. Skip lowercase.
            and s(m[3])                                                 # Rejects import thunks (_imp_ prefix), C++ destructors, and string literals
            and (name := m[3])                                          # Define "name" from the result, for more filtering
            and name not in {"_local_stdio_scanf_options", "vfscanf_l", "vsscanf_l"}
            and not internal.search(name)                               # Filter out the internal symbols, and 3 others giving trouble.
            and (r := f'    {name}{" DATA"*(m[2].upper() in "BDRSV")}') # Format the export line; data symbols (B/D/R/S/V types) get a trailing "DATA".
        ))

@feature('gendef')
@after_method('apply_link')
def add_def(self):
    lt = self.link_task
    if not self.env.llvm_nm:
        return  # Return if llvm-nm isn't found

    obj_files = [f for f in lt.inputs if f.suffix() == '.o']
    d = self.path.find_or_declare(f'{self.target}.def') # Find/Declare the .def output path in the build directory
    self.create_task('nm_def', obj_files, d)            # Create the task
    lt.dep_nodes.append(d)                              # Tell the linker to wait for the .def file before linking

    lt.env.append_unique('LINKFLAGS', [f'/DEF:{d.abspath()}','/DLL']) # Pass the .def file to the linker, with /DLL.


#!/usr/bin/python
import csv

# ctags -o /tmp/atag.csv --fields=+afiKkmnsSzt libs/ardour/ardour/export_format_specification.h 

f = open('/tmp/atag.csv', 'rb')
reader = csv.reader(f, delimiter='\t')
for row in reader:
    if len(row) > 7:
        print row[7]
